#include "coldplug2.h"

/**
 * Function supplementing 'echo add > uevent'
 *
 * @1 File to be written
 *
 * Returns: 0 on success, -1 on failure.
 */
static int coldplug2_uevent(const char *filename) {
	FILE *fp;
	size_t written;

	fp = fopen(filename, "w");
	if (fp == NULL)
		return -1;
	written = fwrite("add\n", 4, 1, fp);
	fclose(fp);

	return (written == 4) ? 0 : -1;
}

/**
 * Recurrent attempt to find all 'uevent' files. Does not
 * follow symlinks.
 *
 * @1 Base directory
 * 
 * Returns: 0 on success, -1 on failure.
 */
static int coldplug2_uevent_lookup(const char *basepath) {
	char entrypath[PATH_MAX];
	DIR *dir;
	int rv;
	size_t baselen;
	struct dirent *dent;
	struct stat statbuf;

	dir = opendir(basepath);
	if (dir == NULL)
		return -1;

	baselen = strlen(basepath);
	if (baselen > PATH_MAX)
		return -1;

	strcpy(entrypath, basepath);
	while ((dent = readdir(dir)) != NULL) {
		if (dent->d_name[0] == '.')
			continue;

		strncat(entrypath, "/", PATH_MAX-baselen);
		strncat(entrypath, dent->d_name, PATH_MAX-baselen-1);

		if (strcmp(dent->d_name, "uevent") == 0) {
			/* If it is an uevent, trigger it. */
			coldplug2_uevent(entrypath);
		} else {
			if (lstat(entrypath, &statbuf) == 0 && S_ISDIR(statbuf.st_mode)) {
				rv = coldplug2_uevent_lookup(entrypath);
				if (rv != 0)
					return rv;
			}
		}

		entrypath[baselen] = '\0';
	}
	closedir(dir);

	return 0;
}

int main(int argc, char *argv[]) {
	char basedir[PATH_MAX];
	int i;
	size_t sysfs_path_len;

	if (argc < 2) {
		/*
		 * No arguments? Then scan entire sysfs.
		 */
		coldplug2_uevent_lookup(SYSFS_PATH);
	} else {
		/* If user asked for help, provide it. */
		if (strcmp(argv[1], "-h") == 0) {
			printf("Usage: %s [<directory1> [<directory2> [...]]]\n\n", argv[0]);
			printf("Execution without any directories specified will scan entire /sys.\n");
			printf("You may specify list of directories, such as 'bus/pci' and 'block' instead.\n");
			return 1;
		}

		/*
		 * Specification of what to actually query was given.
		 */
		sysfs_path_len = strlen(SYSFS_PATH);
		if (sysfs_path_len > PATH_MAX)
			return -1;
		strcpy(basedir, SYSFS_PATH);
		strncat(basedir, "/", PATH_MAX-sysfs_path_len);

		for (i = 1; i < argc; i++) {
			strncat(basedir, argv[i], PATH_MAX-sysfs_path_len-1);
			coldplug2_uevent_lookup(SYSFS_PATH);
			basedir[sysfs_path_len+1] = '\0';
		}
	}

	return 0;
}
